#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_alchemy.zmi.test module

This module defines SQLAlchemy engines test components.
"""

from html import escape
from itertools import tee

from sqlalchemy.exc import DatabaseError, ResourceClosedError
from zope.interface import Interface, alsoProvides
from zope.schema import Text

from pyams_alchemy.engine import get_user_session
from pyams_alchemy.interfaces import IAlchemyEngineUtility, IAlchemyManager, \
    MANAGE_SQL_ENGINES_PERMISSION
from pyams_alchemy.zmi import AlchemyManagerEnginesTable
from pyams_form.ajax import ajax_form_config
from pyams_form.button import Buttons, handler
from pyams_form.field import Fields
from pyams_form.interfaces.form import IAJAXFormRenderer, IGroup
from pyams_layer.interfaces import IPyAMSLayer
from pyams_skin.interfaces.viewlet import IFormFooterViewletManager
from pyams_skin.schema.button import CloseButton, SubmitButton
from pyams_skin.viewlet.help import AlertMessage
from pyams_table.column import GetAttrColumn
from pyams_table.interfaces import IColumn, IValues
from pyams_utils.adapter import ContextRequestViewAdapter, adapter_config
from pyams_utils.interfaces.data import IObjectData
from pyams_utils.list import boolean_iter
from pyams_viewlet.viewlet import EmptyViewlet, ViewContentProvider, viewlet_config
from pyams_zmi.form import AdminModalAddForm, FormGroupSwitcher
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.table import ButtonColumn, InnerTableAdminView, Table


__docformat__ = 'restructuredtext'

from pyams_alchemy import _  # pylint: disable=ungrouped-imports


@adapter_config(name='test',
                required=(IAlchemyManager, IAdminLayer, AlchemyManagerEnginesTable),
                provides=IColumn)
class AlchemyManagerTestColumn(ButtonColumn):
    """SQLAlchemy manager name column"""

    label = _("Test engine")
    status = 'secondary'

    permission = MANAGE_SQL_ENGINES_PERMISSION
    href = 'test-sql-engine.html'
    modal_target = True

    weight = 50


class IAlchemyEngineTestQuery(Interface):
    """SQLAlchemy engine test interface"""

    query = Text(title=_("SQL query"),
                 description=_("Enter valid SQL code; WARNING: query will NOT be committed!!"),
                 required=True)


class IAlchemyEngineTestButtons(Interface):
    """SQLAlchemy engine test buttons"""

    test = SubmitButton(name='test',
                        title=_("Execute query"))

    close = CloseButton(name='close',
                        title=_("Close"))


@ajax_form_config(name='test-sql-engine.html',
                  context=IAlchemyEngineUtility, layer=IPyAMSLayer,
                  permission=MANAGE_SQL_ENGINES_PERMISSION)
class AlchemyEngineTestForm(AdminModalAddForm):
    # pylint: disable=abstract-method
    """SQLAlchemy engine test form"""

    @property
    def title(self):
        """Form title getter"""
        return self.request.localizer.translate(_("SQL engine: {}")).format(self.context.name)

    modal_class = 'modal-max'

    fields = Fields(Interface)
    buttons = Buttons(IAlchemyEngineTestButtons)

    @handler(IAlchemyEngineTestButtons['test'])
    def handle_test(self, action):
        """Query test button handler"""
        data, errors = self.extract_data()
        if not errors:
            self.finished_state.update({
                'action': action,
                'changes': data
            })


@adapter_config(name='query',
                required=(IAlchemyEngineUtility, IAdminLayer, AlchemyEngineTestForm),
                provides=IGroup)
class AlchemyEngineQueryGroup(FormGroupSwitcher):
    """SQLAlchemy engine query group"""

    legend = _("Test SQLAlchemy engine")
    fields = Fields(IAlchemyEngineTestQuery)
    switcher_mode = 'always'

    def update_widgets(self, prefix=None):
        super().update_widgets(prefix)
        query = self.widgets.get('query')
        if query is not None:
            query.add_class('height-100')
            query.widget_css_class = "editor height-100px"
            query.object_data = {
                'ams-filename': 'query.sql'
            }
            alsoProvides(query, IObjectData)


@adapter_config(name='test',
                required=(IAlchemyEngineUtility, IAdminLayer, AlchemyEngineTestForm),
                provides=IAJAXFormRenderer)
class AlchemyEngineTestActionRenderer(ContextRequestViewAdapter):
    """SQLAlchemy engine test action renderer"""

    def render(self, changes):
        """AJAX form renderer"""
        if not changes:
            return None
        test_view = AlchemyEngineTestView(self.context, self.request, self.view)
        try:
            test_view.update()
        except DatabaseError as exc:
            return {
                'status': 'error',
                'messages': [
                    escape('{!r}'.format(exc))
                ]
            }
        return {
            'status': 'success',
            'content': {
                'target': '#{}-test-target'.format(self.view.id),
                'html': test_view.render()
            },
            'closeForm': False
        }


@viewlet_config(name='engine-test.target',
                context=IAlchemyEngineUtility, layer=IAdminLayer, view=AlchemyEngineTestForm,
                manager=IFormFooterViewletManager, weight=20)
class AlchemyEngineTestTarget(EmptyViewlet):
    """SQLAlchemy engine test target"""

    def render(self):
        """Viewlet renderer"""
        return '<div id="{}-test-target" class="hidden"></div>'.format(self.view.id)


class SQLColumn(GetAttrColumn):
    """SQL column"""

    default_value = '--'

    def __init__(self, context, request, table, col):
        super().__init__(context, request, table)
        self.__name__ = col
        self.header = col
        self.attr_name = col

    def get_value(self, obj):
        if obj is not None and self.attr_name is not None:
            value = getattr(obj, self.attr_name)
            if value is None:
                value = escape('<NULL>')
            return value
        return self.default_value


class AlchemyEngineTestTable(Table):
    """SQLAlchemy engine test table"""

    batch_size = 999
    sort_on = None

    results = ()

    css_classes = {
        'table': 'table table-striped table-hover table-xs datatable'
    }

    @property
    def object_data(self):
        """Table data getter"""
        result = super().object_data.copy()
        result.update({
            'col-reorder': True,
            'buttons': 'all',
            'order': []
        })
        return result

    def update(self):
        data = self.form.finished_state.get('changes')  # pylint: disable=no-member
        if 'query' in data:
            session = get_user_session(self.context.name, join=False, twophase=False,
                                       use_zope_extension=False)
            try:
                self.results = list(session.execute(data['query']))
            except ResourceClosedError:
                pass
        super().update()

    def init_columns(self):
        columns = []
        has_values, values = boolean_iter(self.values)
        if has_values:
            rows, _values = tee(values)
            try:
                row = next(rows)
            except StopIteration:
                pass
            else:
                for col in row.keys():
                    column = SQLColumn(self.context, self.request, self, col)
                    columns.append(column)
        self.columns = columns

    def render(self):
        result = super().render()
        if len(self.rows) == self.batch_size:
            alert = AlertMessage(self.context, self.request, self, None)
            alert.status = 'warning'
            alert.css_class = 'p-1 mb-0'
            # pylint: disable=protected-access
            alert._message = self.request.localizer.translate(
                _("Request results contains more than {batch} records; only the "
                  "first records are displayed.")).format(batch=self.batch_size)
            alert.update()
            result += alert.render()
        return result


@adapter_config(required=(IAlchemyEngineUtility, IAdminLayer, AlchemyEngineTestTable),
                provides=IValues)
class AlchemyEngineTestTableValues(ContextRequestViewAdapter):
    """SQLAlchemy engine test table values"""

    @property
    def values(self):
        """Alchemy engine test table values getter"""
        yield from self.view.results


class AlchemyEngineTestView(InnerTableAdminView, ViewContentProvider):
    """SQLAlchemy engine test view"""

    title = _("SQLAlchemy engine test")
    table_class = AlchemyEngineTestTable
    table_label = _("SQL query results")

    def __init__(self, context, request, form):
        super().__init__(context, request)
        self.table.form = form



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_alchemy.interfaces module

This module defines mail package interfaces.
"""

from zope.annotation import IAttributeAnnotatable
from zope.container.constraints import contains
from zope.container.interfaces import IContainer
from zope.interface import Interface
from zope.schema import Bool, Choice, Int, TextLine

from pyams_security.interfaces import IContentRoles
from pyams_security.schema import PrincipalsSetField
from pyams_utils.interfaces import ENCODINGS_VOCABULARY_NAME

from pyams_alchemy import _


#
# SQLAlchemy manager permissions
#

MANAGE_SQL_ENGINES_PERMISSION = 'pyams.ManageAlchemyEngines'
'''Permission used to manage SQLAlchemy engines'''

SQL_MANAGER_ROLE = 'pyams.AlchemyManager'
'''SQL engines manager role'''


#
# SQLAlchemy engine interfaces
#

REQUEST_SESSION_KEY = 'pyams_alchemy.session'

ALCHEMY_ENGINES_VOCABULARY = 'pyams_alchemy.engines'


class IAlchemyEngineUtility(IAttributeAnnotatable):
    """SQLAlchemy engine definition interface"""

    name = TextLine(title=_("Engine name"),
                    description=_("Keep empty if this engine is the default engine..."),
                    required=False,
                    default='')

    dsn = TextLine(title=_('DSN'),
                   description=_('RFC-1738 compliant URL for the database connection'),
                   required=True,
                   default=u'sqlite://')

    echo = Bool(title=_('Echo SQL?'),
                description=_("Log all SQL statements to system logger"),
                required=True,
                default=False)

    use_pool = Bool(title=_("Use connections pool?"),
                    description=_("If 'no', collections pooling will be disabled"),
                    required=True,
                    default=True)

    pool_size = Int(title=_("Pool size"),
                    description=_("SQLAlchemy connections pool size"),
                    required=False,
                    default=25)

    pool_recycle = Int(title=_("Pool recycle time"),
                       description=_("SQLAlchemy connection recycle time (-1 for none)"),
                       required=False,
                       default=-1)

    echo_pool = Bool(title=_("Echo pool?"),
                     description=_("Log all pool checkouts/checkins to system logger?"),
                     required=True,
                     default=False)

    encoding = Choice(title=_('Encoding'),
                      required=True,
                      vocabulary=ENCODINGS_VOCABULARY_NAME,
                      default='utf-8')

    convert_unicode = Bool(title=_('Convert Unicode'),
                           required=True,
                           default=False)

    twophase = Bool(title=_("Two-phases commit?"),
                    description=_("Disable this option if two-phases commits should be "
                                  "disabled (for SQLite for example)"),
                    required=True,
                    default=True)

    def get_engine(self, use_pool=True):
        """Get SQLAlchemy engine"""

    def clear_engine(self):
        """Remove inner volatile attributes when utility properties are modified"""


ALCHEMY_MANAGER_NAME = 'SQL connections'


class IAlchemyManager(IContainer):
    """SQLAlchemy engines container"""

    contains(IAlchemyEngineUtility)


class IAlchemyManagerRoles(IContentRoles):
    """SQLAlchemy manager roles"""

    sql_managers = PrincipalsSetField(title=_("SQL managers"),
                                      role_id=SQL_MANAGER_ROLE,
                                      required=False)


ALCHEMY_CONVERTERS_VOCABULARY = 'pyams_alchemy.converters'


class IAlchemyConverter(Interface):
    """Utility interface provided to convert an SQLAlchemy result proxy"""

    def convert(self, rows, **kwargs):
        """Convert provided rows to format provided by converter"""



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_alchemy.task module

This module defines a PyAMS_scheduler task which can be used to schedule
any SQL command execution.
"""

import sys
import traceback

from sqlalchemy.exc import ResourceClosedError, SQLAlchemyError
from zope.schema.fieldproperty import FieldProperty

from pyams_alchemy.engine import get_user_session
from pyams_alchemy.interfaces import IAlchemyConverter
from pyams_alchemy.task.interfaces import IAlchemyTask
from pyams_scheduler.interfaces.task import TASK_STATUS_EMPTY, TASK_STATUS_ERROR, TASK_STATUS_OK
from pyams_scheduler.task import Task
from pyams_utils.factory import factory_config
from pyams_utils.registry import get_utility


__docformat__ = 'restructuredtext'

from pyams_alchemy import _  # pylint: disable=ungrouped-imports


@factory_config(IAlchemyTask)
class AlchemyTask(Task):
    """SQLAlchemy task"""

    label = _("SQL query")
    icon_class = 'fas fa-database'

    session_name = FieldProperty(IAlchemyTask['session_name'])
    query = FieldProperty(IAlchemyTask['query'])
    output_format = FieldProperty(IAlchemyTask['output_format'])

    def run(self, report, **kwargs):  # pylint: disable=unused-argument
        """Run SQL query task"""
        try:
            session = get_user_session(self.session_name, join=False, twophase=False,
                                       use_zope_extension=False)
            try:
                report.write('SQL query output\n'
                             '================\n')
                report.write('SQL query: \n    {}\n\n'.format(
                    self.query.replace('\r', '').replace('\n', '\n    ')))
                results = session.execute(self.query)
                converter = get_utility(IAlchemyConverter, name=self.output_format)
                result = converter.convert(results)
                session.commit()
                report.write('SQL output ({} records):\n\n'.format(results.rowcount))
                report.write(result)
                return TASK_STATUS_OK, result
            except ResourceClosedError:
                report.write('SQL query returned no result.\n')
                return TASK_STATUS_EMPTY, None
        except SQLAlchemyError:
            etype, value, tb = sys.exc_info()  # pylint: disable=invalid-name
            report.write('\n\n'
                         'An SQL error occurred\n'
                         '=====================\n')
            report.write(''.join(traceback.format_exception(etype, value, tb)))
            session.abort()
            return TASK_STATUS_ERROR, None



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS Alchemy.include module

This module is used for Pyramid integration
"""

import logging
import re

from pyams_alchemy.engine import ConnectionCleanerThread
from pyams_alchemy.interfaces import MANAGE_SQL_ENGINES_PERMISSION, SQL_MANAGER_ROLE
from pyams_security.interfaces import ADMIN_USER_ID, SYSTEM_ADMIN_ROLE
from pyams_security.interfaces.base import ROLE_ID


__docformat__ = 'restructuredtext'

from pyams_alchemy import _  # pylint: disable=ungrouped-imports


LOGGER = logging.getLogger('PyAMS (SQLAlchemy)')


def include_package(config):
    """Pyramid package include"""

    # add translations
    config.add_translation_dirs('pyams_alchemy:locales')

    # add SQLAlchemy management permissions
    config.register_permission({
        'id': MANAGE_SQL_ENGINES_PERMISSION,
        'title': _("Manage SQL engines properties")
    })

    # upgrade system manager roles
    config.upgrade_role(SYSTEM_ADMIN_ROLE,
                        permissions={
                            MANAGE_SQL_ENGINES_PERMISSION
                        })

    # register new roles
    config.register_role({
        'id': SQL_MANAGER_ROLE,
        'title': _("SQL engines manager (role)"),
        'permissions': {
            MANAGE_SQL_ENGINES_PERMISSION
        },
        'managers': {
            ADMIN_USER_ID,
            ROLE_ID.format(SYSTEM_ADMIN_ROLE)
        }
    })

    # package scan
    ignored = []
    try:
        import pyams_zmi  # pylint: disable=import-outside-toplevel,unused-import
    except ImportError:
        ignored.append(re.compile(r'pyams_alchemy\..*\.zmi\.?.*').search)

    try:
        import pyams_scheduler  # pylint: disable=import-outside-toplevel,unused-import
    except ImportError:
        ignored.append('pyams_alchemy.task')

    config.scan(ignore=ignored)

    timeout = config.registry.settings.get('pyams_alchemy.cleaner.timeout', '300')
    if timeout and (timeout.lower() not in ('off', 'false', '0', 'disabled')):
        LOGGER.info("Starting SQLAlchemy connections management thread...")
        cleaner_thread = ConnectionCleanerThread(int(timeout))
        cleaner_thread.daemon = True
        cleaner_thread.start()



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_alchemy.loader module

This module defines a base class which can be used to duplicate entities from one session
to another one.
"""

__docformat__ = 'restructuredtext'

import logging

from sqlalchemy.orm import make_transient

from pyams_alchemy.engine import get_user_session


LOGGER = logging.getLogger('PyAMS (alchemy)')


class DataLoader:
    """SQLAlchemy data loader

    This utility class is used to migrate entities from a given connection
    to another one.
    WARNING: actually, given entities must share the same schema name!!!
    """

    def __init__(self, source, target, entities):
        """
        Initialize data loader

        :param source: name of registered source engine
        :param target: name of registered target engine
        :param entities: list of migrated entities
        """
        self.source_session = get_user_session(source)
        self.target_session = get_user_session(target)
        self.entities = entities

    def run(self):
        """Run data loader"""
        source = self.source_session
        target = self.target_session
        for entity in self.entities:
            LOGGER.info('Loading entity {0!r}'.format(entity))
            for record in source.query(entity):
                source.expunge(record)
                make_transient(record)
                target.add(record)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_alchemy.zmi.engine module

This module defines SQL engines management components.
"""

from pyramid.events import subscriber
from zope.copy import copy
from zope.interface import Interface, Invalid, implementer

from pyams_alchemy.interfaces import IAlchemyEngineUtility, IAlchemyManager, \
    MANAGE_SQL_ENGINES_PERMISSION
from pyams_alchemy.zmi import AlchemyManagerEnginesTable
from pyams_form.ajax import ajax_form_config
from pyams_form.field import Fields
from pyams_form.interfaces import DISPLAY_MODE
from pyams_form.interfaces.form import IAJAXFormRenderer, IAddForm, IDataExtractedEvent
from pyams_layer.interfaces import IPyAMSLayer
from pyams_skin.viewlet.actions import ContextAction
from pyams_table.interfaces import IColumn
from pyams_utils.adapter import ContextRequestViewAdapter, adapter_config
from pyams_utils.interfaces.intids import IUniqueID
from pyams_utils.registry import query_utility
from pyams_utils.traversing import get_parent
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import AdminModalAddForm, AdminModalEditForm
from pyams_zmi.helper.event import get_json_table_row_add_callback, \
    get_json_table_row_refresh_callback
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.interfaces.table import ITableElementEditor
from pyams_zmi.interfaces.viewlet import IToolbarViewletManager
from pyams_zmi.table import ActionColumn, TableElementEditor
from pyams_zmi.utils import get_object_label


__docformat__ = 'restructuredtext'

from pyams_alchemy import _  # pylint: disable=ungrouped-imports


class IAlchemyEngineAddForm(IAddForm):
    """SQLAlchemy engine add form interface"""


@subscriber(IDataExtractedEvent, form_selector=IAlchemyEngineAddForm)
def handle_new_engine_data_extraction(event):
    """Handle new engine data"""
    name = event.data['name'] or ''
    engine = query_utility(IAlchemyEngineUtility, name=name)
    if engine is not None:
        event.form.widgets.errors += (Invalid(_("An SQLAlchemy engine is already "
                                                "registered with this name!")),)


@adapter_config(required=(IAlchemyManager, IAdminLayer, IAlchemyEngineAddForm),
                provides=IAJAXFormRenderer)
class AlchemyEngineAddFormRenderer(ContextRequestViewAdapter):
    """Alchemy engine add form AJAX renderer"""

    def render(self, changes):
        """AJAX form renderer"""
        if not changes:
            return None
        manager = get_parent(self.context, IAlchemyManager)
        return {
            'callbacks': [
                get_json_table_row_add_callback(manager, self.request,
                                                AlchemyManagerEnginesTable, changes)
            ]
        }


#
# Alchemy engine add form
#

@viewlet_config(name='add-sql-engine.menu',
                context=IAlchemyManager, layer=IAdminLayer, view=AlchemyManagerEnginesTable,
                manager=IToolbarViewletManager, weight=10,
                permission=MANAGE_SQL_ENGINES_PERMISSION)
class AlchemyEngineAddMenu(ContextAction):
    """Alchemy engine add menu"""

    status = 'success'
    icon_class = 'fas fa-plus'
    label = _("Add SQL engine")

    href = 'add-sql-engine.html'
    modal_target = True


@implementer(IAlchemyEngineAddForm)
class AlchemyEngineBaseAddFormMixin:
    """Alchemy engine base add form mixin"""

    @property
    def title(self):
        translate = self.request.localizer.translate
        return '<small>{}</small><br />{}'.format(
            get_object_label(self.context, self.request, self),
            translate(_("New SQL engine")))


@ajax_form_config(name='add-sql-engine.html',
                  context=IAlchemyManager, layer=IPyAMSLayer,
                  permission=MANAGE_SQL_ENGINES_PERMISSION)
class AlchemyEngineAddForm(AlchemyEngineBaseAddFormMixin, AdminModalAddForm):  # pylint: disable=abstract-method
    """SQLAlchemy engine add form"""

    legend = _("New engine properties")

    fields = Fields(IAlchemyEngineUtility)
    content_factory = IAlchemyEngineUtility

    def add(self, obj):
        oid = IUniqueID(obj).oid
        self.context[oid] = obj


#
# Alchemy engine clone form
#

@adapter_config(name='clone',
                required=(IAlchemyManager, IAdminLayer, AlchemyManagerEnginesTable),
                provides=IColumn)
class AlchemyEngineCloneColumn(ActionColumn):
    """SQLAlchemy engine clone column"""

    hint = _("Clone SQL engine")
    icon_class = 'far fa-clone'

    href = 'clone-sql-engine.html'

    weight = 100


@ajax_form_config(name='clone-sql-engine.html',
                  context=IAlchemyEngineUtility, layer=IPyAMSLayer,
                  permission=MANAGE_SQL_ENGINES_PERMISSION)
class AlchemyEngineCloneForm(AlchemyEngineBaseAddFormMixin, AdminModalAddForm):
    """SQLAlchemy engine clone form"""

    legend = _("Clone SQL connection")

    fields = Fields(IAlchemyEngineUtility).select('name')

    def create(self, data):
        return copy(self.context)

    def add(self, obj):
        oid = IUniqueID(obj).oid
        self.context.__parent__[oid] = obj


#
# Alchemy engine edit form
#

@adapter_config(required=(IAlchemyEngineUtility, IAdminLayer, Interface),
                provides=ITableElementEditor)
class AlchemyEngineElementEditor(TableElementEditor):
    """SQLAlchemy engines table element editor"""


@ajax_form_config(name='properties.html',
                  context=IAlchemyEngineUtility, layer=IPyAMSLayer,
                  permission=MANAGE_SQL_ENGINES_PERMISSION)
class AlchemyEngineEditForm(AdminModalEditForm):
    """SQLAlchemy engine properties edit form"""

    @property
    def title(self):
        translate = self.request.localizer.translate
        manager = query_utility(IAlchemyManager)
        return '<small>{}</small><br />{}'.format(
            get_object_label(manager, self.request, self),
            translate(_("SQL engine: {}")).format(self.context.name))

    legend = _("SQL engine properties")

    fields = Fields(IAlchemyEngineUtility)

    def update_widgets(self, prefix=None):
        super().update_widgets(prefix)
        name = self.widgets.get('name')
        if name is not None:
            name.mode = DISPLAY_MODE


@adapter_config(required=(IAlchemyEngineUtility, IAdminLayer, AlchemyEngineEditForm),
                provides=IAJAXFormRenderer)
class AlchemyEngineEditFormAJAXRenderer(ContextRequestViewAdapter):
    """SQLAlchemy engine edit form AJAX renderer"""

    def render(self, changes):
        """AJAX result renderer"""
        if not changes:
            return None
        manager = get_parent(self.context, IAlchemyManager)
        return {
            'callbacks': [
                get_json_table_row_refresh_callback(manager, self.request,
                                                    AlchemyManagerEnginesTable, self.context)
            ]
        }



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_alchemy.manager module

This module defines an SQLAlchemy "manager", which is just a
container for SQLAlchemy engines definitions.
"""

from zope.container.folder import Folder
from zope.interface import implementer

from pyams_alchemy.interfaces import IAlchemyManager, IAlchemyManagerRoles
from pyams_security.interfaces import IDefaultProtectionPolicy, IRolesPolicy
from pyams_security.property import RolePrincipalsFieldProperty
from pyams_security.security import ProtectedObjectMixin, ProtectedObjectRoles
from pyams_utils.adapter import ContextAdapter, adapter_config
from pyams_utils.factory import factory_config


__docformat__ = 'restructuredtext'


@factory_config(IAlchemyManager)
@implementer(IDefaultProtectionPolicy)
class AlchemyManager(ProtectedObjectMixin, Folder):
    """SQLAlchemy engines container"""


@implementer(IAlchemyManagerRoles)
class AlchemyManagerRoles(ProtectedObjectRoles):
    """SQLAlchemy manager roles"""

    sql_managers = RolePrincipalsFieldProperty(IAlchemyManagerRoles['sql_managers'])


@adapter_config(required=IAlchemyManager,
                provides=IAlchemyManagerRoles)
def alchemy_manager_roles_adapter(context):
    """SQLAlchemy manager roles adapter"""
    return AlchemyManagerRoles(context)


@adapter_config(name='alchemy_roles',
                required=IAlchemyManager,
                provides=IRolesPolicy)
class AlchemyManagerRolesPolicy(ContextAdapter):
    """SQLAlchemy manager roles policy"""

    roles_interface = IAlchemyManagerRoles
    weight = 20
